/* 
C language dll: GSPy
Original file name: CustomPythoncDLL.c
Created by: Nick Martin, GoldSim Technology Group
Creation Date: 21 May 2015
Revised by: Ryan Roper, GoldSim Technology Group
Last Edited: 13 April 2018

License: FreeBSD License (reproduced below)

Purpose:

This file provides a direct interface to a custom, pure python module.

Python must be installed on this machine so that Python.h can be found by
the compiler.
*/

/*
LICENSE

This source code is available for your use under the FreeBSD License, see:
http://directory.fsf.org/wiki?title=License:FreeBSD 

FreeBSD License

Copyright (c) 2015, GoldSim Technology Group LLC
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of the FreeBSD Project.
*/

//#define ENABLEGSPYDIAGNOSTICS

#include <Python.h>
#include "GSPyInterface.h"
#include "ErrorRelay.h"

// Create some pointers in the global scope
static PyObject *pName, *pModule, *pDict, *pFunc;
static PyObject *pArgs, *pValue;
static const char* pyModuleErrorMessage = "Python Module Error! Details were not specified.";

// Below is a key function that intializes the Python interpreter and sets (within this
// running instance of the interpreter) the PYTHONHOME environment variable so that GSPy
// Python dependencies (i.e. Python modules/libraries required by GSPyModule.py) can be
// found. However, GSPy.pyd must first be able to find the user-provided custom GSPyModule.py
// file. For this, the current working directory (i.e. the location of GSPy.pyd) must be
// added to sys.path. This is done after PYTHONHOME is set and after the Python interpreter
// is initialized. Adding "." to sys.path is required, because the interpreter does not, by
// default, look in the working directory for Python files.
//
// In short, for GSPy.pyd to find GSPyModule.py (assuming it is in the same directory), the
// working directory (i.e. ".", relative to GSPy.pyd) must be added to sys.path. For any Python
// modules that GSPyModule.py relies on to be found, PYTHONHOME must be set to point at the
// installation folder of the Python distribution (i.e. the location of python3.dll, etc.).
//
// Additionally, GSPy.pyd must be able to find DLL dependencies such as python3.dll. The 
// purpose of setting PYTHONHOME when the Python interpreter is initialized/launched is not
// to make it so GSPy.pyd can find python3.dll and other dependencies. These dependencies are
// searched for when GoldSim first loads GSPy.pyd (before PYTHONHOME is ever set). Therefore,
// the user will need to set an environment variable, GSPY (retrieved during the call to
// initializePythonInterpreter()), which should then be added by the user to the system PATH.
void initializePythonInterpreter()
{
	// Get the string value of the GSPY environment variable
	char* pPath;
	pPath = getenv(GSPYENVVAR);

	// If a string value for the GSPY environment variable was found, set PYTHONHOME to this value
	if (pPath != NULL)
	{
		// Get the size of pPath and add 1 (to be used in converting the char* to a wchar_t*)
		size_t size = strlen(pPath) + 1;

		// Create a wchar_t*, which will be used to set PYTHONHOME (using Py_SetPythonHome)
		wchar_t* envvar_wchar = new wchar_t[size];

		// The address of the following variable must be provided as an argument to mbstowcs_s
		// (the function used to copy pPath, a char*, to envvar_wchar, a wchar_t*)
		size_t outSize;

		// Copy pPath (a char*) to envvar_wchar (a wchar_t*)
		mbstowcs_s(&outSize, envvar_wchar, size, pPath, size - 1);

		// Set PYTHONHOME for the current instance of the Python interpreter (this is essential
		// for GSPy - specifically GSPyModule.py - to find Python module/library dependencies)
		Py_SetPythonHome(envvar_wchar);
	}

	// Initialize the Python interpreter
	Py_Initialize();

	// Once the Python interpreter has been initialized, retrieve the PATH environment
	// variable and append the current working directory (i.e. ".", relative to the location
	// of GSPy.pyd) so that Python module/library dependencies can be found. Note, this only
	// modifies the PATH environment variable for the current Python interpreter instance. 
	PyObject* sysPath = PySys_GetObject((char*)"path");
	PyList_Append(sysPath, PyUnicode_FromString("."));
}

static const char* diagnosticsFileName = "GSPyDiagnostics.txt";
void tempFileIOFunction(char* s)
{
	FILE *diagFile = fopen(diagnosticsFileName, "a");
	fputs(s, diagFile);
	fclose(diagFile);
}

// DLL interface for GoldSim
void GSPyInterface ( int methodID, int * status, double * inargs, double * outargs ) {
    // initialize a few c variables to be used later.
    int RetStatus1 = 0;
    int RetStatus2 = 0;
    int RetStatus3 = 0;
    int RetStatus4 = 0;
	int RetStatus5 = 0;
    double VersionReturn = -1.0;
    int PyStatus = 0;
    // set the status to bad to start.
    *status = XF_FAILURE;

    // The first thing to do is to set up the interaction with the Python interpreter.
	// GoldSim may want to keep this dll loaded. To avoid repeatedly initializing Python,
	// always check the current status with Py_IsInitialized() (where == 0 is false; <> 0
	// is true). Py_Initialize() and Py_Finalize() are available from <Python.h>. Continue
	// based on the selected method.
    switch ( methodID ) {
        case XF_INITIALIZE:
            // Check if Python is initialized
            PyStatus = Py_IsInitialized();
            // If not, then initialize
			#ifdef ENABLEGSPYDIAGNOSTICS
			tempFileIOFunction((char*)"XF_INITIALIZE Called.\n");
			#endif
            if ( PyStatus == 0 )
			{
				#ifdef ENABLEGSPYDIAGNOSTICS
				tempFileIOFunction((char*)"Python interpreter initialized in XF_INITIALIZE.\n");
				#endif
				initializePythonInterpreter();
            }

            // This C function checks for an optional initialization method in GSPyModule.py
			// and calls that method if it exists
            RetStatus1 = InitRoutines();

			// Return value of 1 indicates an error occurred
            if ( RetStatus1 == 1 )
			{
                // Set status to failure with message
				*status = XF_FAILURE_WITH_MSG;

				// Copy Python module error to outargs
				CopyMsgToOutputs(pyModuleErrorMessage, outargs);
					
				//char* pChar = (char*)outargs;
				//char* ErrMsg = (char*)MAKELONG(MAKEWORD(pChar[0], pChar[1]), MAKEWORD(pChar[2], pChar[3]));
				//printf("Returned from InitRoutines(): %s\n", ErrMsg);

				// Clean up Python, if necessary
                PyStatus = Py_IsInitialized();
                if ( PyStatus != 0 )
				{
                    Py_Finalize();
                }
            }
            else
			{
                *status = XF_SUCCESS;
            }
            break;

        case XF_REP_VERSION:
			// Check if Python is initialized
			PyStatus = Py_IsInitialized();
			// If not, then initialize
			#ifdef ENABLEGSPYDIAGNOSTICS
			tempFileIOFunction((char*)"XF_REP_VERSION Called.\n");
			#endif
            if ( PyStatus == 0 )
			{
				#ifdef ENABLEGSPYDIAGNOSTICS
				tempFileIOFunction((char*)"Python interpreter initialized in XF_REP_VERSION.\n");
				#endif
				initializePythonInterpreter();
            }

            // This C function calls a Python module method to get the version
            VersionReturn = ReturnPyModuleVersion();

            // VersionReturn has a default negative value to catch Python
			// at runtime compilation issues.
            if ( VersionReturn > 0.0 )
			{
                outargs[0] = VersionReturn;
                *status = XF_SUCCESS;
            }
            else
			{
				// Set status to failure with message
                *status = XF_FAILURE_WITH_MSG;

				// Return the standard XF_REP_VERSION error message
                ReturnErrorMsg( XF_REP_VERSION, outargs );

                // Clean up Python, if necessary
                PyStatus = Py_IsInitialized();
                if ( PyStatus != 0 )
				{
                    Py_Finalize();
                }
            }
            break;

        case XF_REP_ARGUMENTS:
			// Check if Python is initialized
            PyStatus = Py_IsInitialized();
			// If not, then initialize
			#ifdef ENABLEGSPYDIAGNOSTICS
			tempFileIOFunction((char*)"XF_REP_ARGUMENTS Called.\n");
			#endif
            if ( PyStatus == 0 )
			{
				#ifdef ENABLEGSPYDIAGNOSTICS
				tempFileIOFunction((char*)"Python interpreter initialized in XF_REP_ARGUMENTS.\n");
				#endif
				initializePythonInterpreter();
            }

			// This C function calls a Python module method to get the number
			// of arguments expected from GSPy/GoldSim.
            RetStatus3 = NumInputsExpected();

			// A negative result indicates an error
            if ( RetStatus3 < 0 )
			{
				// Set status to failure with message
                *status = XF_FAILURE_WITH_MSG;

				// Return the standard XF_REP_ARGUMENTS error message
                ReturnErrorMsg( XF_REP_ARGUMENTS, outargs );
            }
            else
			{
                *status = XF_SUCCESS;
            }

			// Store number of expected inputs to outargs
            outargs[0] = (double) RetStatus3;

			// This C function calls a Python module method to get the number
			// of outputs expected from the Python module calculations.
            RetStatus4 = NumOutputsToProvide();

			// A negative result indicates an error
            if ( RetStatus4 < 0 )
			{
				// Set status to failure with message
                *status = XF_FAILURE_WITH_MSG;

				// Return the standard XF_REP_ARGUMENTS error message
                ReturnErrorMsg( XF_REP_ARGUMENTS, outargs );
            }
            else
			{
                *status = XF_SUCCESS;
            }

			// Store number of expected outputs to outargs
            outargs[1] = (double) RetStatus4; 

			// Check if any errors occurred while getting the number
			// of inputs and outputs
            if ( ( RetStatus3 < 0 ) || ( RetStatus4 < 0 ) )
			{
				// Clean up Python, if necessary
                PyStatus = Py_IsInitialized();
                if ( PyStatus != 0 )
				{
                    Py_Finalize();
                }
            }
            break;

        case XF_CALCULATE:
			// Check if Python is initialized
            PyStatus = Py_IsInitialized();
			// If not, then initialize
			#ifdef ENABLEGSPYDIAGNOSTICS
			tempFileIOFunction((char*)"XF_CALCULATE Called.\n");
			#endif
            if ( PyStatus == 0 )
			{
				#ifdef ENABLEGSPYDIAGNOSTICS
				tempFileIOFunction((char*)"Python interpreter initialized in XF_CALCULATE.\n");
				#endif
				initializePythonInterpreter();
            }

			// This C function calls a Python module method to run custom
			// calculations defined in GSPyModule.py
            RetStatus2 = DoCalcsAndReturnValues( inargs, outargs );
            
			// A non-zero return value indicates an error occurred
			if ( RetStatus2 != 0 )
			{
				// Set status to failure with message
                *status = XF_FAILURE_WITH_MSG;
                
				// A negative return value indicates that the Python
				// module generated an error message
				if (RetStatus2 < 0)
				{
					// Copy the Python module error message to outargs
					CopyMsgToOutputs(pyModuleErrorMessage, outargs);
				}
				else
				{
					// Return the standard XF_CALCULATE error message
					ReturnErrorMsg(XF_CALCULATE, outargs);
				}
                
				// Clean up Python, if necessary
                PyStatus = Py_IsInitialized();
                if ( PyStatus != 0 )
				{
                    Py_Finalize();
                }
            }
            else
			{
                *status = XF_SUCCESS;
            }
            break;

        case XF_CLEANUP:
			// Check if Python is initialized
            PyStatus = Py_IsInitialized();
			// If not, then initialize
			#ifdef ENABLEGSPYDIAGNOSTICS
			tempFileIOFunction((char*)"XF_CLEANUP Called.\n");
			#endif
            if ( PyStatus == 0 )
			{
				#ifdef ENABLEGSPYDIAGNOSTICS
				tempFileIOFunction((char*)"Python interpreter initialized in XF_CLEANUP.\n");
				#endif
				initializePythonInterpreter();
            }

			// By default, set the status to success
            *status = XF_SUCCESS;

			// This C function checks for an optional finalize method in GSPyModule.py
			// and calls that method if it exists
			RetStatus5 = WrapUpSimulation();

			// Return value of 1 indicates an error occurred
			if (RetStatus5 == 1)
			{
				// Set status to failure with message
				*status = XF_FAILURE_WITH_MSG;
				
				// Copy Python module error to outargs
				CopyMsgToOutputs(pyModuleErrorMessage, outargs);
				
				// Clean up Python, if necessary
				PyStatus = Py_IsInitialized();
				if (PyStatus != 0)
				{
					Py_Finalize();
				}
			}
			else
			{
				*status = XF_SUCCESS;
			}

			// Clean up Python, if necessary
            PyStatus = Py_IsInitialized();
            if ( PyStatus != 0 )
			{
                Py_Finalize();
            }
            break;

        default:
			// Check if Python is initialized
            PyStatus = Py_IsInitialized();
			// If not, then initialize
			#ifdef ENABLEGSPYDIAGNOSTICS
			tempFileIOFunction((char*)"Default Called.\n");
			#endif
            if ( PyStatus == 0 )
			{
				#ifdef ENABLEGSPYDIAGNOSTICS
				tempFileIOFunction((char*)"Python interpreter initialized in Default.\n");
				#endif
				initializePythonInterpreter();
            }

			// Set status to failure with message
            *status = XF_FAILURE_WITH_MSG;

			// Return a default error message
            ReturnErrorMsg( 5, outargs );

			// Clean up Python, if necessary
            PyStatus = Py_IsInitialized();
            if ( PyStatus != 0 )
			{
                Py_Finalize();
            }
            break;
    }
}

// Error messaging function
void ReturnErrorMsg(int errcode, double *outargs)
{
	switch (errcode)
	{
		case(XF_PYCOMPILE_ERR):
			CopyMsgToOutputs(Error_in_Python_Compilation, outargs);
			break;
		case(XF_INITIALIZE):
			CopyMsgToOutputs(Error_in_GSPy_DLL_in_INITIALIZATION, outargs);
			break;
		case(XF_CALCULATE):
			CopyMsgToOutputs(Error_in_GSPy_DLL_in_CALCULATION, outargs);
			break;
		case(XF_REP_VERSION):
			CopyMsgToOutputs(Error_in_GSPy_DLL_in_VERSION_REPORT, outargs);
			break;
		case(XF_REP_ARGUMENTS):
			CopyMsgToOutputs(Error_in_GSPy_DLL_in_ARGUMENTS_REPORT, outargs);
			break;
		case(XF_CLEANUP):
			CopyMsgToOutputs(Error_in_GSPy_DLL_in_CLEANUP, outargs);
			break;
		default:
			CopyMsgToOutputs(Error_in_GSPy_DLL_Unsupported, outargs);
			break;
	}
}

// Calls the initialization method (if defined) in the Python module
int InitRoutines()
{
	// Import the Python module, if not already imported
	if (pModule == NULL)
	{
		pName = PyUnicode_DecodeFSDefault(PYMODNAME);
		pModule = PyImport_Import(pName);
	}

	// If Python module is still not loaded, return with non-zero value
	if (pModule == NULL)
	{
		PyErr_Print();
		fprintf(stderr, "Failed to load \"%s\"\n", PYMODNAME);
		return 1;
	}
	
	if (pModule != NULL) // Python module successfully loaded
	{
		// Check if an initialization function is defined
		if (PyObject_HasAttrString(pModule, PYMODINITIALIZE))
		{
			// Get the initialization function from the Python module
			pFunc = PyObject_GetAttrString(pModule, PYMODINITIALIZE);
		}
		// Optional Initialization function not defined in Python module
		else
		{ 
			printf("Note: Optional \"%s\" function not defined in Python module\n", PYMODINITIALIZE);
			return 0;
		}

		if (pFunc && PyCallable_Check(pFunc)) // Check that pFunc is valid and callable
		{ 
			// Call initialization function in Python module
			pValue = PyObject_CallObject(pFunc, NULL);

			if (pValue != NULL) // Check if valid result obtained from initialization function
			{
				// Check for Py_None return value (indicating an error)
				if (pValue == Py_None)
				{
					// Return of 1 indicates fatal error
					return 1;
				}

				// Decrement reference counter for pValue
				Py_DECREF(pValue);
			}
			// Valid results were not obtained from initialization function
			else
			{
				Py_DECREF(pFunc);
				PyErr_Print();
				fprintf(stderr, "Call to Python function \"%s\" failed!\n", PYMODINITIALIZE);
				return 1;
			}
		}
		// Python object pFunc is not valid and/or callable
		else
		{
			if (PyErr_Occurred())
				PyErr_Print();
			fprintf(stderr, "Cannot find Python function \"%s\"\n", PYMODINITIALIZE);
		}

		// Decrement reference counter for pFunc, if necessary
		Py_XDECREF(pFunc);
	}
	// Python module could not be loaded
	else
	{
		fprintf(stderr, "Could not load module \"%s\"\n", PYMODNAME);
		return 1;
	}
	
	return 0;
}

// Calls the calculation method in the Python module
int DoCalcsAndReturnValues(double *inargs, double *outargs)
{
	// Import the Python module, if not already imported
	if (pModule == NULL)
	{
		pName = PyUnicode_DecodeFSDefault(PYMODNAME);
		pModule = PyImport_Import(pName);
	}

	// Retrieve the number of input and output arguments
	int size_inargs = NumInputsExpected();
	int size_outargs = NumOutputsToProvide();
	
	if (pModule != NULL) // Python module successfully loaded
	{
		// Get the calculation function from the Python module
		pFunc = PyObject_GetAttrString(pModule, PYMODCALCULATIONS);
		
		if (pFunc && PyCallable_Check(pFunc)) // Check that pFunc is valid and callable
		{
			// Create a Python tuple object to store arguments
			pArgs = PyTuple_New(size_inargs);
			
			// Loop over inargs and store values in pArgs
			for (int i = 0; i < size_inargs; i++)
			{
				PyTuple_SetItem(pArgs, i, PyFloat_FromDouble(inargs[i]));
			}

			// Call the calculation function from the Python module
			pValue = PyObject_CallObject(pFunc, pArgs);

			// Decrement reference counter for pArgs
			Py_DECREF(pArgs);

			if (pValue != NULL) // Check if valid results obtained from calculation function
			{
				// Check for Py_None return value (indicating an error)
				if (pValue == Py_None)
				{
					// Return of -1 indicates fatal error in Python module
					return -1;
				}

				// Loop over calculation results and store values in outargs
				for (int i = 0; i < size_outargs; i++)
				{
					outargs[i] = PyFloat_AsDouble(PyTuple_GetItem(pValue, i));
				}

				// Decrement reference counter for pValue
				Py_DECREF(pValue);
			}
			// Valid results were not obtained from calculation function
			else
			{
				Py_DECREF(pFunc);
				PyErr_Print();
				fprintf(stderr, "Call to Python function \"%s\" failed!\n", PYMODCALCULATIONS);
				return 1;
			}
		}
		// Python object pFunc is not valid and/or callable
		else
		{
			if (PyErr_Occurred())
				PyErr_Print();
			fprintf(stderr, "Cannot find Python function \"%s\"\n", PYMODCALCULATIONS);
		}
		//Py_XDECREF(pFunc);
		Py_CLEAR(pFunc);
		if (PyCallable_Check(pFunc))
		{
			printf("Warning: pFunc is still callable after Py_CLEAR.\n");
		}
	}
	// Python module could not be loaded
	else
	{
		fprintf(stderr, "Could not load module \"%s\"\n", PYMODNAME);
		return 1;
	}

	return 0;
}

// Calls a method in the Python module that returns the version of the module
double ReturnPyModuleVersion(void)
{
	// Default version value is -1.0 (if returned, this implies an error occurred)
	double version = -1.0;

	// Import the Python module, if not already imported
	if (pModule == NULL)
	{
		pName = PyUnicode_DecodeFSDefault(PYMODNAME);
		pModule = PyImport_Import(pName);
	}

	if (pModule != NULL) // Python module successfully loaded
	{
		// Get the version report function from the Python module
		pFunc = PyObject_GetAttrString(pModule, PYMODGETVERSION);
		
		if (pFunc && PyCallable_Check(pFunc)) // Check that pFunc is valid and callable
		{
			// Call the version report function from the Python module
			pValue = PyObject_CallObject(pFunc, NULL);

			if (pValue != NULL) // Check if valid result obtained from version report function
			{
				// Convert from PyFloat to C double
				version = PyFloat_AsDouble(pValue);
				printf("Version number reported from Python module: %1.2f\n", version);

				// Decrement reference counter for pValue
				Py_DECREF(pValue);
			}
			// Valid result was not obtained from version report function
			else
			{
				Py_DECREF(pFunc);
				PyErr_Print();
				fprintf(stderr, "Call to Python function \"%s\" failed!\n", PYMODGETVERSION);
				return -1;
			}
		}
		// Python object pFunc is not valid and/or callable
		else
		{
			if (PyErr_Occurred())
				PyErr_Print();
			fprintf(stderr, "Cannot find Python function \"%s\"\n", PYMODGETVERSION);
		}

		// Decrement reference counter for pFunc, if necessary
		Py_XDECREF(pFunc);
	}
	// Python module could not be loaded
	else
	{
		fprintf(stderr, "Could not load module \"%s\"\n", PYMODNAME);
		return -1;
	}

	return version;
}

// Calls a method in the Python module that returns the expected number of inputs
int NumInputsExpected(void)
{
	// Default number of inputs is -1 (if returned, this implies an error occurred)
	int numinputs = -1;

	// Import the Python module, if not already imported
	if (pModule == NULL)
	{
		pName = PyUnicode_DecodeFSDefault(PYMODNAME);
		pModule = PyImport_Import(pName);
	}

	if (pModule != NULL) // Python module successfully loaded
	{
		// Get the function that reports the number of inputs from the Python module
		pFunc = PyObject_GetAttrString(pModule, PYMODGETNUMINPUTS);

		if (pFunc && PyCallable_Check(pFunc)) // Check that pFunc is valid and callable
		{
			// Call the inputs reporting function from the Python module
			pValue = PyObject_CallObject(pFunc, NULL);

			if (pValue != NULL) // Check if valid result obtained from inputs reporting function
			{
				// Convert from PyLong to C long int
				numinputs = PyLong_AsLong(pValue);
				printf("Number of inputs reported from Python module: %d\n", numinputs);

				// Decrement reference counter for pValue
				Py_DECREF(pValue);
			}
			// Valid result was not obtained from inputs reporting function
			else
			{
				Py_DECREF(pFunc);
				PyErr_Print();
				fprintf(stderr, "Call to Python function \"%s\" failed!\n", PYMODGETNUMINPUTS);
				return -1;
			}
		}
		// Python object pFunc is not valid and/or callable
		else
		{
			if (PyErr_Occurred())
				PyErr_Print();
			fprintf(stderr, "Cannot find Python function \"%s\"\n", PYMODGETNUMINPUTS);
		}

		// Decrement reference counter for pFunc, if necessary
		Py_XDECREF(pFunc);
	}
	// Python module could not be loaded
	else
	{
		fprintf(stderr, "Could not load module \"%s\"\n", PYMODNAME);
		return -1;
	}

	return numinputs;
}

// Calls a method in the Python module that returns the expected number of outputs
int NumOutputsToProvide(void)
{
	// Default number of outputs is -1 (if returned, this implies an error occurred)
	int numoutputs = -1;

	// Import the Python module, if not already imported
	if (pModule == NULL)
	{
		pName = PyUnicode_DecodeFSDefault(PYMODNAME);
		pModule = PyImport_Import(pName);
	}

	if (pModule != NULL) // Python module successfully loaded
	{
		// Get the function that reports the number of outputs from the Python module
		pFunc = PyObject_GetAttrString(pModule, PYMODGETNUMOUTPUTS);

		if (pFunc && PyCallable_Check(pFunc)) // Check that pFunc is valid and callable
		{
			// Call the outputs reporting function from the Python module
			pValue = PyObject_CallObject(pFunc, NULL);

			if (pValue != NULL) // Check if valid result obtained from outputs reporting function
			{
				// Convert from PyLong to C long int
				numoutputs = PyLong_AsLong(pValue);
				printf("Number of outputs reported from Python module: %d\n", numoutputs);

				// Decrement reference counter for pValue
				Py_DECREF(pValue);
			}
			// Valid result was not obtained from outputs reporting function
			else
			{
				Py_DECREF(pFunc);
				PyErr_Print();
				fprintf(stderr, "Call to Python function \"%s\" failed!\n", PYMODGETNUMOUTPUTS);
				return -1;
			}
		}
		// Python object pFunc is not valid and/or callable
		else
		{
			if (PyErr_Occurred())
				PyErr_Print();
			fprintf(stderr, "Cannot find Python function \"%s\"\n", PYMODGETNUMOUTPUTS);
		}

		// Decrement reference counter for pFunc, if necessary
		Py_XDECREF(pFunc);
	}
	// Python module could not be loaded
	else
	{
		fprintf(stderr, "Could not load module \"%s\"\n", PYMODNAME);
		return -1;
	}

	return numoutputs;
}

// Calls the finalize method (if defined) in the Python module
int WrapUpSimulation(void)
{
	// Import the Python module, if not already imported
	if (pModule == NULL)
	{
		pName = PyUnicode_DecodeFSDefault(PYMODNAME);
		pModule = PyImport_Import(pName);
	}

	if (pModule != NULL) // Python module successfully loaded
	{
		// Check if a finalize function is defined
		if (PyObject_HasAttrString(pModule, PYMODFINALIZE))
		{
			// Get the finalize function from the Python module
			pFunc = PyObject_GetAttrString(pModule, PYMODFINALIZE);
		}
		// Optional Finalize function not defined in Python module
		else
		{
			printf("Note: Optional \"%s\" function not defined in Python module\n", PYMODFINALIZE);
			return 0;
		}

		if (pFunc && PyCallable_Check(pFunc)) // Check that pFunc is valid and callable
		{
			// Call finalize function in Python module
			pValue = PyObject_CallObject(pFunc, NULL);

			if (pValue != NULL) // Check if valid result obtained from finalize function
			{
				// Check for Py_None return value (indicating an error)
				if (pValue == Py_None)
				{
					// Return of 1 indicates fatal error
					return 1;
				}

				// Decrement reference counter for pValue
				Py_DECREF(pValue);
			}
			// Valid results were not obtained from finalize function
			else
			{
				Py_DECREF(pFunc);
				PyErr_Print();
				fprintf(stderr, "Call to Python function \"%s\" failed!\n", PYMODFINALIZE);
			}
		}
		// Python object pFunc is not valid and/or callable
		else
		{
			if (PyErr_Occurred())
				PyErr_Print();
			fprintf(stderr, "Cannot find Python function \"%s\"\n", PYMODFINALIZE);
		}
		
		// Decrement reference counter for pFunc, if necessary
		Py_XDECREF(pFunc);
	}
	// Python module could not be loaded
	else
	{
		fprintf(stderr, "Could not load module \"%s\"\n", PYMODNAME);
	}

	// Decrement the reference count for pName (the Python module name)
	Py_XDECREF(pName);

	// Decrement the reference count for pModule (the Python module object)
	Py_XDECREF(pModule);

	return 0;
}

// Stub function to report Python compilation error
void PyCompileError(void)
{

}

// ========== THE CODE BELOW DEFINES A PYTHON API ========== //

// The functions defined below can be called from a pure Python module. Simply import GSPy
// and then call a method like this: GSPy.setErrorMessage(<input argument>)

// Python API function that can be called from GSPyModule.py to set the value of pyModuleErrorMessage.
// This is used for relaying error messages to GSPy/GoldSim from the Python module.
static PyObject* GSPy_setErrorMessage(PyObject *self, PyObject *msg)
{
	// Convert the expected Python string to a char* and store in pyModuleErrorMessage
	int ok = PyArg_ParseTuple(msg, "s", &pyModuleErrorMessage);

	if (ok) // If parsing of input argument was successful
	{
		// Note that the important work has already been done (i.e. storing the error
		// message in pyModuleErrorMessage so it's available to GSPy/GoldSim to display
		// to the user. The message, below, is only seen when running TestGSPy.exe
		printf("An error message was relayed from the Python module\n");
	}
	
	// Must return a Python 'None' since return cannot be void
	return Py_None;
}

// Basic (and unnecessary) Python API function that was added to confirm that multiple
// functions can be added to GSPy and called from GSPyModule.py.
static PyObject* GSPy_add(PyObject *self, PyObject *args)
{
	// This function requires/expects two doubles as input arguments
	double num1, num2;

	// Convert the expected Python floats to C doubles and store in num1 and num2
	int ok = PyArg_ParseTuple(args, "dd", &num1, &num2);

	if (ok) // If parsing of input arguments was successful
	{
		// Calculate the sum
		double result = num1 + num2;

		// Return value as a Python float
		return PyFloat_FromDouble(result);
	}

	// Must return a Python 'None' since return cannot be void
	return Py_None;
}

// Add Python API functions to the module methods list
static PyMethodDef GSPy_methods[] = {
	{"setErrorMessage", (PyCFunction) GSPy_setErrorMessage, METH_VARARGS, "Allows a Python module to return an error message"},
	{"add", (PyCFunction) GSPy_add, METH_VARARGS, "Adds two numbers together"},
	{NULL, NULL, 0, NULL}
};

// Specify properties of the GSPy module (e.g. the name "GSPy") and
// add the methods list to the module
static struct PyModuleDef GSPyModule = {
	PyModuleDef_HEAD_INIT,
	"GSPy",			// Name of module
	NULL,
	-1,				// Module keeps state in global variables
	GSPy_methods	// Add methods list
};

// Create the Python API module
PyMODINIT_FUNC PyInit_GSPy(void)
{
	return PyModule_Create(&GSPyModule);
}

//EOF